/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.UUID;

public class MemoryStorage extends AbstractTreeStorage {
    private UUID root = null;
    private Map<UUID, String> toId = new HashMap<UUID, String>();
    private Map<UUID, Collection<UUID>> toChildren = new LinkedHashMap<UUID, Collection<UUID>>();
    private Map<UUID, Collection<Attribute>> toAttributes = new LinkedHashMap<UUID, Collection<Attribute>>();
    private Map<UUID, HashMap<String, ByteArrayOutputStream>> toResourсes = new LinkedHashMap<UUID, HashMap<String, ByteArrayOutputStream>>();

    public void emulateDeltaCreate(UUID parentUUId, UUID uuid, String type,
            String id, Collection<Attribute> attributes) {
        toId.put(uuid, id);
        toChildren.put(uuid, new HashSet<UUID>());
        toAttributes.put(uuid, attributes);
        if (parentUUId == null) {
            root = uuid;
        } else {
            toChildren.get(parentUUId).add(uuid);
        }
        getDeltaTarget().created(uuid, false);
    }

    public void emulateDeltaDelete(UUID uuid) {
        toId.remove(uuid);
        toAttributes.remove(uuid);
        Collection<UUID> childrens = toChildren.remove(uuid);
        for (UUID puuid : toChildren.keySet()) {
            if (toChildren.get(puuid).contains(uuid)) {
                toChildren.get(puuid).remove(uuid);
            }
        }
        for (UUID cuuid : childrens) {
            emulateDeltaDelete(cuuid);
        }
        getDeltaTarget().deleted(uuid, false);
    }

    public void emulateDeltaMove(UUID uuid, UUID newParent) {
        UUID oldParent = null;
        for (UUID puuid : toChildren.keySet()) {
            if (toChildren.get(puuid).contains(uuid)) {
                oldParent = puuid;
            }
        }
        assert oldParent != null;

        toChildren.get(oldParent).remove(uuid);
        toChildren.get(newParent).add(uuid);
        getDeltaTarget().moved(uuid, oldParent, newParent, false);
    }

    public void emulateDeltaChanged(UUID uuid, Collection<Attribute> attributes) {
        toAttributes.put(uuid, attributes);
        getDeltaTarget().changed(uuid, false);
    }

    @Override
    public UUID getRootNode() {
        return root;
    }

    @Override
    public String getNodeId(UUID uuid) {
        return toId.get(uuid);
    }

    @Override
    public void createNode(UUID parentUUId, UUID uuid, String id,
            Collection<Attribute> attrs) {
        toId.put(uuid, id);
        toChildren.put(uuid, new HashSet<UUID>());
        if (attrs == null) {
            attrs = new HashSet<Attribute>();
        }
        toAttributes.put(uuid, new HashSet<Attribute>(attrs));
        if (parentUUId == null) {
            root = uuid;
        } else {
            toChildren.get(parentUUId).add(uuid);
        }
        getDeltaTarget().created(uuid, true);
    }

    @Override
    public void remove(UUID uuid) {
        toId.remove(uuid);
        toAttributes.remove(uuid);
        Collection<UUID> toDelete = new ArrayList<UUID>(toChildren.get(uuid));
        for(UUID duuid: toDelete) {
            remove(duuid);
        }
        // remove from parent
        for (UUID puid : toChildren.keySet()) {
            if (toChildren.get(puid).contains(uuid)) {
                toChildren.get(puid).remove(uuid);
            }
        }
        toChildren.remove(uuid);
        getDeltaTarget().deleted(uuid, true);
    }

    @Override
    public Collection<Attribute> getAttributes(UUID uuid) {
        return Collections.unmodifiableCollection(toAttributes.get(uuid));
    }

    @Override
    public Collection<UUID> getChildren(UUID uuid) {
        if (toChildren.get(uuid) == null) {
            return null;
        }
        return Collections.unmodifiableCollection(toChildren.get(uuid));
    }

    @Override
    public UUID getParent(UUID uuid) {
        for (UUID i : toChildren.keySet()) {
            if (toChildren.get(i).contains(uuid)) {
                return i;
            }
        }
        return null;
    }

    @Override
    public boolean canLoad(UUID uuid) {
        return toId.containsKey(uuid);
    }

    @Override
    public void changeNodeId(UUID uuid, String id) {
        toId.put(uuid, id);
        getDeltaTarget().changed(uuid, true);
    }

    @Override
    public void moveNodeTo(UUID node, UUID parent) {
        UUID oldParent = getParent(node);
        if(oldParent != null)
            toChildren.get(oldParent).remove(node);
        toChildren.get(parent).add(node);
        getDeltaTarget().moved(node, oldParent, parent, true);
    }

    @Override
    public void saveAttributes(UUID uuid, Collection<Attribute> attributes) {
        toAttributes.put(uuid, new HashSet<Attribute>(attributes));
        getDeltaTarget().changed(uuid, true);
    }

    @Override
    public String getNodeResourcePath(UUID uuId) {
        throw new UnsupportedOperationException();
    }

    @Override
    public InputStream getResourceContents(UUID uuid, String resName) {
        // TODO check stream
        byte[] data = getResourceData(uuid, resName);
        if (data == null) {
            return null;
        }
        return new ByteArrayInputStream(data);
    }

    @Override
    public void setResourceContents(UUID uuid, String resName, InputStream in, Object data) {
        if (!toResourсes.containsKey(uuid)) {
            toResourсes.put(uuid, new HashMap<String, ByteArrayOutputStream>());
        }
        Map<String, ByteArrayOutputStream> r = toResourсes.get(uuid);
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        r.put(resName, out);
        byte[] buffer = new byte[1024];
        int len;
        try {
            while ((len = in.read(buffer)) != -1) {
                out.write(buffer, 0, len);
            }
            getDeltaTarget().changed(uuid, true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    
    private byte[] getResourceData(UUID uuid, String resName) {
        if(!isResourceExists(uuid, resName)) {
            return null;
        }
        Map<String, ByteArrayOutputStream> nodeResources = toResourсes.get(uuid);
        ByteArrayOutputStream resData = nodeResources.get(resName);
        return resData.toByteArray();
    }

    @Override
    public boolean isResourceExists(UUID uuid, String resName) {
        Map<String, ByteArrayOutputStream> nodeResources = toResourсes.get(uuid);
        if (nodeResources == null) {
            return false;
        }
        ByteArrayOutputStream resData = nodeResources.get(resName);
        if (resData == null) {
            return false;
        }
        return true;
    }

    @Override
    public void deleteResource(UUID uuid, String resName) {
        toResourсes.get(uuid).remove(resName);
        getDeltaTarget().changed(uuid, true);
    }

    @Override
    public String[] listResources(UUID uuid) {
        return toResourсes.get(uuid).keySet().toArray(new String[0]);
    }

    @Override
    public void createStorage(String name) {
    }

    @Override
    public void deleteStorage() {
    }

    @Override
    public String getName() {
        return toString();
    }
}
