/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.core;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.NodeType;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.nodetypes.TestPurpose;

public class TreeDB {
    public static final String ORDER_ATTRIBUTE = "_order";
    public static final String INDEX_ATTRIBUTE = "_index";
    private List<ITreeChangeListener> listeners = new CopyOnWriteArrayList<ITreeChangeListener>();
    private TreeNode root;
    private Map<UUID, TreeNode> cache = new HashMap<UUID, TreeNode>();
    private ITreeStorage storage;
    private IUUIDGenerator uuidGenerator;
    private TreeLogic tlogic;
    private HashMap<String, ArrayList<UUID>> namecache = new HashMap<String, ArrayList<UUID>>();
    private HashMap<UUID, String> backnamecache = new HashMap<UUID, String>(); 
    private Map<String, NodeType> defaultTypes = new HashMap<String, NodeType>();
    {
        NodeType dbroot = new NodeType("DBRoot", TreeDB.this);
        dbroot.setClass(com.unitesk.requality.core.TreeNode.class);
        defaultTypes.put("DBRoot", dbroot);
        NodeType nodetype = new NodeType("NodeType", TreeDB.this);
        nodetype.setClass(com.unitesk.requality.nodetypes.NodeType.class);
        defaultTypes.put("NodeType", nodetype);
        NodeType node = new NodeType("Node", TreeDB.this);
        node.setClass(com.unitesk.requality.core.TreeNode.class);
        defaultTypes.put("Node", node);
    };

    public TreeLogic getTreeLogic(){
        if(tlogic == null)
            tlogic = new TreeLogic();
        return tlogic;
    }
    
    private TreeDB() {
        changelisteners = new HashMap<UUID, INodeChangeListener>();
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // API methods
    //
    // //////////////////////////////////////////////////////////////////////////
    public TreeNode createChildNode(UUID pUUId, String id, String type) {
        UUID uuid = uuidGenerator.nextUUId();
        return createChildNode(uuid, pUUId, id, type);
    }

    public TreeNode createChildNode(UUID uuid, UUID pUUId, String id,
            String type) {
        if (getNode(pUUId).findChild(id) != null) {
            throw new IllegalArgumentException("Child with id: "
                    + " exists in " + getNode(pUUId).getQualifiedId());
        }
        storage.createNode(pUUId, uuid, type, id);
        return getNode(uuid);
    }

    public void removeNode(TreeNode node) {
        // Remove children first
        while (node.sizeChildren() > 0) {
            removeNode(node.getChild(0));
        }
        UUID uuid = node.getUUId();
        storage.remove(uuid);
        node.setDisposed();
    }

    public void moveNode(TreeNode node, TreeNode newParent) {
        storage.moveNodeTo(node.getUUId(), newParent.getUUId());
    }

    void setId(TreeNode node, String id) {
        if (!node.getId().equals(id)) {
            storage.changeNodeId(node.getUUId(), id);
            node.fireNodeAttributeChange("_id");
        }
    }

    /**
     * @return root node of TreeDB
     */
    public TreeNode getRootNode() {
        if (root == null) {
            UUID rootUUID;
            if((rootUUID = storage.getRootNode())==null) { // root doesn't exists  
                return null;
            }
            root = getNode(rootUUID);
        }
        return root;
    }
    
    /**
     * Method to update cached requirement name.
     * @param node requirement to update name for
     */
    public synchronized void updateRequirementName(TreeNode node){
        if(!(node instanceof Requirement))
            return;
        Requirement r = (Requirement)node;
        ArrayList<UUID> list;
        String oldName = "";
        //Update cache - remove old name
        if(backnamecache.containsKey(r.getUUId())){
            oldName = backnamecache.get(r.getUUId());
            if(!oldName.equals("") && (oldName.equals(r.getName())))
                return;
            list = namecache.get(oldName);
            list.remove(r.getUUId());
            namecache.put(oldName, list);
        }
        //if we have new name is not set, stop  
        if(r.getName() == null)
            return;
        //if we have name, add it to cache
        list = namecache.get(r.getName());
        if(list == null)
            list = new ArrayList<UUID>();
        list.add(r.getUUId());
        namecache.put(r.getName(), list);
        backnamecache.put(r.getUUId(), r.getName());
    }
    
    /**
     * Method to get Array of TreeNodes with same User Visible Id
     * @param path User Visible Name to find path with
     */
    public synchronized String[] getNodesByUFName(String path, List<String> types){
        ArrayList<TreeNode> curnodes = new ArrayList<TreeNode>();
        StringTokenizer st = new StringTokenizer(path, "/");
        String nodeId = st.nextToken();
        //if we have node with name nodeId in requirement then use it
        TreeNode rnode = getNode("/Requirements/"+nodeId);
        if(rnode != null)
            curnodes.add(rnode);
        //if cache of Requiremens names contains nodeId
        //then try to get appropriate node
        if (namecache.containsKey(nodeId)) {
            ArrayList<UUID>nodes = namecache.get(nodeId);
            if(nodes.size() < 1)
                return null;
            for(UUID uid : nodes)
                curnodes.add(getNode(uid));
        }
        if (curnodes.size() == 0) {
            return null;
        }
        // assume that all nodes in curnodes can
        // contain node with UFName. nullsize - number
        // of empty elements in array
        int csize = curnodes.size(), nullsize = 0;
        TreeNode node;
        while (st.hasMoreTokens()) {
            nodeId = st.nextToken();
            nullsize = 0;
            for(int i = csize-1; i >= 0; i--){
                node = curnodes.get(i);
                //find child in node
                if(node != null)
                    node = node.findChild(nodeId);
                curnodes.set(i, node);
                //if we have no child with this name,
                
                if(node == null){
                    nullsize++;
                    continue;
                }
            }
            if(nullsize == csize)
                return null;
        }
        if(types.size() > 0)
            for(int i = 0; i < csize; i++){
                TreeNode n = curnodes.get(i);
                if(n != null && !types.contains(n.getType())){
                    curnodes.set(i, null);
                    nullsize++;
                }
            }
        //array to return 
        String[]ret = new String[csize-nullsize];
        int j = 0;
        //get array of UUIDs strings to return
        int num = curnodes.size();
        for (int i = 0; i < num; i++) {
            node = curnodes.get(i);
            if (node != null)
                ret[j++] = node.getUUId().toString();
        }
        return ret;
    }
    private Map<UUID,INodeChangeListener> changelisteners;
    @SuppressWarnings("unchecked")
    public synchronized TreeNode getNode(UUID uuid) {
    	TreeNode node = cache.get(uuid); 
        if (node != null) {
            return node;
        }
        if (storage.canLoad(uuid)) {
            //code below was written to deal with the problem of
            //loops in some case and to add change listeners 
            // for linked Nodes. 
            TreeNode n = createFromStorage(uuid, storage.getNodeType(uuid));
            //fix for event loop
            cache.put(n.getUUId(), n);
            //update of name cache for Requirements
            //Using for UFName supporting.
            if(n.getType().equals(Requirement.TYPE_NAME))
                updateRequirementName(n);
            //Get links if they are contained in 
            //current node
            Attribute a = n.getAttribute(TreeNode.ATTR_LINKS);
            if(a != null){
                ArrayList<String>links = (ArrayList<String>) a.getValue();
                for(String link : links){
                    //Link with * - hidden links
                    //we need to deal with them in
                    //some situations.
                    if(link.startsWith("*"))
                        link = link.substring(1);
                    //link contains UUID, so we use it here
                    UUID uid = UUID.fromString(link.substring(0, link.indexOf('_')));
                    INodeChangeListener ncl = n.getChangeListener();
                    if(ncl == null)
                        continue;
                    if(!n.getUUId().equals(uid)){
                        try{
                            getNode(uid).addChangeListener(ncl);
                        }
                        catch(RuntimeException re){
                            changelisteners.put(uid, ncl);
                            //Node not found
                        }
                    }else{
                        n.addChangeListener(ncl);
                    }
                }
            }
            if(changelisteners.containsKey(uuid)){
                n.addChangeListener(changelisteners.get(uuid));
                changelisteners.remove(uuid);
            }
            return n;
        } else {
            throw new RuntimeException("node not found: " + uuid);
        }
    }

    public TreeNode getNode(String qId) {
        if (qId.equals("/")) {
            return getRootNode();
        }
        StringTokenizer st = new StringTokenizer(qId, "/");
        TreeNode cur = getRootNode();
        while (st.hasMoreTokens()) {
            String nodeId = st.nextToken();
            cur = cur.findChild(nodeId);
            if (cur == null) {
                return null;
            }
        }
        return cur;
    }

    public void saveAttributes(UUID uuid, Collection<Attribute> attributes) {
        storage.saveAttributes(uuid, attributes);
    }

    public void setResourceContent(TreeNode node, String name, InputStream in, Object data) {
        storage.setResourceContents(node.getUUId(), name, in, data);
        //update Requirement name 
        if(node.getType().equals(Requirement.TYPE_NAME))
            updateRequirementName(node);
    }

    public void deleteResource(TreeNode node, String name) {
        storage.deleteResource(node.getUUId(), name);
    }

    public boolean isResourceExists(TreeNode node, String name) {
        return storage.isResourceExists(node.getUUId(), name);
    }

    public String getName() {
        return storage.getName();
    }

    // //////////////////////////////////////////////////////////////////////////
    //
    // Transaction API methods
    //
    // //////////////////////////////////////////////////////////////////////////
    private ArrayList<ITransactionListener> transactionListeners = new ArrayList<ITransactionListener>();

    public void startTransaction(String name) {
        if (storage instanceof TransactionStorage) {
            throw new IllegalStateException("transaction \"" + getName()
                    + "\"is started");
        }
        TransactionStorage tStorage = new TransactionStorage(storage,
                transactionListeners);
        tStorage.setName(name);
        tStorage.setDeltaTarget(storageDeltaListener);
        storage.setDeltaTarget(nullDeltaListener);
        storage = tStorage;
    }

    public TransactionStorage commit() {
        if (!(storage instanceof TransactionStorage)) {
            throw new IllegalStateException("transaction not started");
        }
        TransactionStorage tStorage = (TransactionStorage) storage;
        tStorage.apply();
        tStorage.setDeltaTarget(nullDeltaListener);
        storage = tStorage.getStorage();
        storage.setDeltaTarget(storageDeltaListener);
        return tStorage;
    }

    public void rollback() {
        if (!(storage instanceof TransactionStorage)) {
            throw new IllegalStateException("transaction not started");
        }
        TransactionStorage tStorage = (TransactionStorage) storage;
        tStorage.rollback(tStorage);
        tStorage.setDeltaTarget(nullDeltaListener);
        storage = tStorage.getStorage();
        storage.setDeltaTarget(storageDeltaListener);
//        for (ITransactionListener tl : transactionListeners) {
//            tl.undo(tStorage);
//        }
    }

    public void addTransactionListener(ITransactionListener l) {
        transactionListeners.add(l);
    }

    public void removeTransactionListener(ITransactionListener l) {
        transactionListeners.remove(l);
    }

    // ////////////////////////////////////////////////////////////////////////
    //
    // Types operations
    //
    // ////////////////////////////////////////////////////////////////////////
    public void addNodeType(String name, Class<? extends TreeNode> clazz) {
        NodeType nt = new NodeType(name, TreeDB.this);
        nt.setClass(clazz);
        defaultTypes.put(name, nt);
    }

    public void addNodeType(Class<? extends TreeNode> clazz) {
        addNodeType(clazz.getSimpleName(), clazz);
    }

    public void addNodeTypes(Map<String, Class<? extends TreeNode>> nts) {
        for (String name : nts.keySet()) {
            addNodeType(name, nts.get(name));
        }
    }

    public Map<String, NodeType> getTypes() {
        Map<String, NodeType> result = new HashMap<String, NodeType>(
                defaultTypes);
        TreeNode typesNode = getNode("/Types");
        if (typesNode != null && typesNode.sizeChildren() > 0) {
            int num = typesNode.sizeChildren();
            for (int i = 0; i < num; i++) {
                NodeType node = (NodeType) typesNode.getChild(i);
                result.put(node.getId(), node);
            }
        }
        return result;
    }

    public void createType(NodeType newType) {
        TreeNode types = getNode("/Types");
        if (types == null) {
            types = getRootNode().createChild("Types", NodeType.TYPE_NAME);
        }
        TreeNode newTypeNode = types.createChild(newType.getName());
        newTypeNode.saveAttributes();
    }

    public NodeType getNodeType(String name) {
        for (NodeType type : getTypes().values()) {
            if (type.getId().equals(name)) {
                return type;
            }
        }
        return null;
    }

    public boolean removeType(String typeName) {
        TreeNode types = getNode("/Types");
        if (types == null) {
            return false;
        }
        TreeNode n = types.findChild(typeName);
        n.delete();
        return true;
    }

    // ////////////////////////////////////////////////////////////////////////
    //
    // TreeDB factories
    //
    // ////////////////////////////////////////////////////////////////////////

    /**
     * Creates db structures and root node
     * 
     * @param name
     * @param storage
     * @param uuidGenerator
     * @return
     */
    public static TreeDB create(String name, ITreeStorage storage,
            IUUIDGenerator uuidGenerator) {
        TreeDB newDB = new TreeDB();
        newDB.uuidGenerator = uuidGenerator;
        storage.createStorage(name);
        storage.setDeltaTarget(newDB.storageDeltaListener);
        newDB.storage = storage;
        // create root node
        UUID uuid = uuidGenerator.nextUUId();
        storage.createNode(null, uuid, "DBRoot", "/");
        newDB.tlogic = new TreeLogic();
        return newDB;
    }

    /**
     * Creates db structures and root node
     * 
     * @param name
     * @param storage
     * @return treeDB object
     */
    public static TreeDB create(String name, ITreeStorage storage) {
        return create(name, storage, new DefaultUUIDGenerator());
    }

    public static TreeDB load(ITreeStorage storage, IUUIDGenerator uuidGenerator) {
        TreeDB newDB = new TreeDB();
        newDB.uuidGenerator = uuidGenerator;
        // FIXME: for test we can use mock target, so we left old one if it is
        // not null. Its look like bad design.
        if (storage.getDeltaTarget() == null) {
            storage.setDeltaTarget(newDB.storageDeltaListener);
        }
        newDB.storage = storage;
        return newDB;
    }

    public static TreeDB load(ITreeStorage storage) {
        return load(storage, new DefaultUUIDGenerator());
    }

    // ////////////////////////////////////////////////////////////////////////
    public boolean addListener(ITreeChangeListener listener) {
        return listeners.add(listener);
    }

    public boolean removeListener(ITreeChangeListener listener) {
        return listeners.remove(listener);
    }

    public List<ITreeChangeListener> getListeners() {
        return Collections.unmodifiableList(listeners);
    }

    private IStorageDeltaListener nullDeltaListener = new IStorageDeltaListener() {
        @Override
        public void moved(UUID uuid, UUID oldParent, UUID newParent,
                boolean internal) {
        }

        @Override
        public void deleted(UUID uuid, boolean internal) {
        }

        @Override
        public void created(UUID uuid, boolean internal) {
        }

        @Override
        public void changed(UUID uuid, boolean internal) {
        }
    };

    private void notifyExternalChange() {
        for (ITransactionListener tl : transactionListeners) {
            tl._dispose();
        }
    }

    private boolean compareNodes(TreeNode node, NodeDesc nodeDesc) {
        if (!node.getUUId().equals(nodeDesc.getUUId()))
            return false;
        UUID nodeParentUUId = node.getParentUUId();
        UUID descParentUUId = nodeDesc.getParentUUId();
        if (nodeParentUUId == null && descParentUUId != null || nodeParentUUId != null && descParentUUId == null)
            return false;
        if (nodeParentUUId != null && descParentUUId != null && !nodeParentUUId.equals(descParentUUId))
            return false;
        if (!node.getId().equals(nodeDesc.getId()))
            return false;
        
        Map<String, Attribute> oldMap = node.getAttributes();
        Map<String, Attribute> newMap = new HashMap<String, Attribute>(nodeDesc.getAttributes());
        for (String attr_key : oldMap.keySet()) {
            Attribute oldAttr = oldMap.get(attr_key);
            Attribute newAttr = null;
            if (newMap.containsKey(attr_key))
                newAttr = newMap.remove(attr_key);
            else
                return false;
            if (oldAttr == null && newAttr != null || oldAttr != null && newAttr == null)
                return false;
            if (oldAttr != null && newAttr != null && !oldAttr.equals(newAttr))
                return false;
        }
        if (!newMap.isEmpty())
            return false;
        
        List<UUID> oldList = node.getChildrenUUIds();
        List<UUID> newList = new ArrayList<UUID>(nodeDesc.getChildrenUUId());
        for (UUID uuid : oldList) {
            if (newList.contains(uuid))
                newList.remove(uuid);
            else
                return false;
        }
        if (!newList.isEmpty())
            return false;
        return true;
    }

    private IStorageDeltaListener storageDeltaListener = new IStorageDeltaListener() {
        @Override
        public void created(UUID uuid, boolean forceInternal) {
            TreeNode curNode = cache.get(uuid);
            NodeDesc newNodeDesc = loadNodeDescription(uuid);
            if (curNode != null) {
                if (!compareNodes(curNode, newNodeDesc)) {
                    throw new IllegalStateException("node with uuid: "
                            + curNode.getUUId() + "(" + curNode.getQualifiedId()
                            + ") exists");
                } else {
                    return;
                }
            }

            curNode = createModelNode(
                    newNodeDesc.getUUId(), 
                    newNodeDesc.getParentUUId(), 
                    newNodeDesc.getId(), 
                    (String)newNodeDesc.getAttributes().get(TreeNode.ATTR_TYPE).getValue(),
                    newNodeDesc.getAttributes().values(),
                    newNodeDesc.getChildrenUUId());
            TreeNode parent = curNode.getParent();
            if (parent != null && !parent.children.contains(uuid)) {
                parent.children.add(uuid); 
            }
            cache.put(uuid, curNode);
            if (curNode.getId().equals("/")) {
                assert curNode.parent == null;
                root = curNode;
            }
            if (!forceInternal) {
                notifyExternalChange();
            }
            for (ITreeChangeListener l : listeners) {
                l.created(curNode, forceInternal);
            }
            //update Requirement name in cache
            if(curNode instanceof Requirement)
                updateRequirementName(curNode);
        }

        private NodeDesc loadNodeDescription(UUID uuid) {
            String typeName = storage.getNodeType(uuid);
            UUID parentUUId;
            String id;
            if (typeName.equals("DBRoot")) {
                id = "/";
                parentUUId = null;
            } else {
                id = storage.getNodeId(uuid);
                parentUUId = storage.getParent(uuid);
            }
            Map<String, Attribute> attributes = new TreeMap<String, Attribute>();
            Collection<Attribute> attrs = storage.getAttributes(uuid);
            for (Attribute a : attrs) {
                attributes.put(a.getKey(), a);
            }
            List<UUID> childrenUUId = new ArrayList<UUID>(
                    storage.getChildren(uuid));
            return new NodeDesc(uuid, parentUUId, id, attributes, childrenUUId);
        }

        @Override
        public void deleted(UUID uuid, boolean internal) {
            TreeNode node = cache.get(uuid);
            //if node is not cache
            if(node == null)
                return;
            NodeDesc nodeDesc = NodeDesc.fromNode(node);
            node.getParent().children.remove(uuid);
            //refresh cache
            if (backnamecache.containsKey(uuid)){
                node.removeAttribute(Requirement.ATTR_NAME);
                updateRequirementName(node);
            }
            node.fireNodeDelete();
            cache.remove(uuid);
            node.setDisposed();
            if (!internal) {
                notifyExternalChange();
            }
            for (ITreeChangeListener l : listeners) {
                l.deleted(nodeDesc, internal);
            }
        }

        @Override
        public void changed(UUID uuid, boolean internal) {
            // Sometimes (e.g, import of existing project)
        	// changed event happens when 
        	// there is no node in the cache yet.
        	// So, use getNode(uuid) instead cache.get(uuid).
        	TreeNode curNode = getNode(uuid);
            String newId = storage.getNodeId(uuid);
            boolean neq = !newId.equals(curNode.id);
            curNode.id = newId;
        	if (curNode instanceof Requirement && neq) {
        	    ((Requirement)curNode).updateTreeNodes((Requirement)curNode, "_id");
        	}
            loadNodeAttributes(uuid, curNode);
            if (!internal) {
                notifyExternalChange();
            }
            fireNodeChange(curNode, internal);
            if(curNode != null)
            	curNode.fireNodeAttributeChange(TreeNode.ATTR_UNDEFINED);
        }

        @Override
        public void moved(UUID nodeUUID, UUID oldParentUUId, UUID newParentUUId,
                boolean internal) {
            TreeNode node = cache.get(nodeUUID);
            TreeNode newParent = cache.get(newParentUUId);
            TreeNode oldParent = cache.get(oldParentUUId);

            node.parent = newParentUUId;
            if(oldParent!=null)
                oldParent.children.remove(nodeUUID);
            newParent.children.add(nodeUUID);

            if (!internal) {
                notifyExternalChange();
            }
            for (ITreeChangeListener l : listeners) {
                l.moved(node, oldParent == null ? node.getParent() : oldParent,
                        internal);
            }
            node.updateTreeNodes(node, "_id");
        }
    };

    private void loadNodeAttributes(UUID uuid, TreeNode node) {
        Collection<Attribute> newAttributes = storage.getAttributes(uuid);
        applyNodeAttributes(node, newAttributes);
    }

    @SuppressWarnings("unchecked")
    private void applyNodeAttributes(TreeNode node, Collection<Attribute> newAttributes) {
        //for new and old Locations of Requirement reset labels
        if (node instanceof Requirement) {
            Requirement req = (Requirement) node;
            String oldName = req.getName();
            String newName = null;
            //get old Locations list
            List<String> oldLocs = new ArrayList<String>(req.getLocationQIds());
            //construct new Locations list
            List<String> newLocs = null;
            for (Attribute attr : newAttributes) {
                if (attr.getKey().equals(Requirement.ATTR_LOCATIONS)) {
                    Object val = attr.getValue();
                    if (val instanceof List) {
                        newLocs = new ArrayList<String>((List<String>) val);
                    }
                } else
                if (attr.getKey().equals("_name")) {
                    newName = (String) attr.getValue();
                }
            }
            //compare lists
            boolean fNewLoc = false;
            if (newLocs != null) {
                for (String newLoc : newLocs) {
                    if (!oldLocs.remove(newLoc)) {
                        fNewLoc = true;
                    }
                }
            }
            //and call affected Locations
            if (fNewLoc || !oldLocs.isEmpty()) {
                //remaining old locations (removed)
                for (String oldLoc : oldLocs) {
                    TreeNode locNode = getNode(oldLoc);
                    if (locNode instanceof Location) {
                        ((Location)locNode).resetLabel();
                    }
                }
                //all other locations
                if (newLocs != null) {
                    for (String newLoc : newLocs) {
                        TreeNode locNode = getNode(newLoc);
                        if (locNode instanceof Location) {
                            ((Location)locNode).resetLabel();
                        }
                    }
                }
            }
            //check, if name was changed
            if (oldName != null || newName != null) {
                if (oldName == null && newName != null || oldName != null && newName == null || !oldName.equals(newName)) {
                    req.updateTreeNodes(req, "_name");
                }
            }
        }
        updateNodeAttributes(node, newAttributes);
    }

    private void updateNodeAttributes(TreeNode node, Collection<Attribute> newAttributes) {
        HashSet<String> removeKeys = new HashSet<String>(
                node.attributes.keySet());
        for (Attribute a : newAttributes) {
            removeKeys.remove(a.getKey());
            node.attributes.put(a.getKey(), a);
        }
        for (String key : removeKeys) {
            node.attributes.remove(key);
        }
        if(node instanceof Requirement)
            updateRequirementName((Requirement)node);
    }

    public IUUIDGenerator getUuidGenerator() {
        return uuidGenerator;
    }

    public ITreeStorage getStorage() {
        return storage;
    }

    private TreeNode createModelNode(UUID uuid, UUID puuid, String id,
            String typeName, Collection<Attribute> attributes, Collection<UUID> children) {
        TreeNode result = null;
        if (defaultTypes.containsKey(typeName)) {
            NodeType nt = defaultTypes.get(typeName);
            try {
                Class<?>[] treeNodeArgs = new Class[] { TreeDB.class,
                        UUID.class };
                result = nt.getTypeClass().getConstructor(treeNodeArgs)
                        .newInstance(TreeDB.this, uuid);
                result.setType(typeName);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
        if (result == null) {
            result = new TreeNode(this, uuid);
            result.setType(typeName);
        }
        result.id = id;
        result.parent = puuid;
        result.children = new ArrayList<UUID>(children);
        cache.put(uuid, result);
        updateNodeAttributes(result, attributes);
        return result;
    }

    private TreeNode createFromStorage(UUID uuid, String typeName) {
        String id;
        UUID puuid;
        if (typeName.equals("DBRoot")) {
            id = "/";
            puuid = null;
        } else {
            id = storage.getNodeId(uuid);
            puuid = storage.getParent(uuid);
        }
        Collection<Attribute> attributes = storage.getAttributes(uuid);
        Collection<UUID> children = storage.getChildren(uuid);
        TreeNode result = createModelNode(uuid, puuid, id, typeName, attributes, children);
        cache.put(uuid, result);
        if (result.getId().equals("/")) {
            assert result.parent == null;
            root = result;
        }
        return result;
    }

    public void fireNodeChange(TreeNode curNode, boolean internal) {
        for (ITreeChangeListener l : listeners) {
            l.changed(curNode, internal);
        }
    }
    private static List<String> typeFilter = new ArrayList<String>();
    /**
     * String for representing of Node's types filter in form of list of string names of types. For ex. {TestPurpose.TYPE_NAME}"
     * @return
     */
    public List<String> getLinkToTypeFilter(String linktype){
        //TODO TreeLogic methods
        if(linktype.equals("TestStepLinks")){
            ArrayList<String> a = new ArrayList<String>();
            a.add(TestPurpose.TYPE_NAME);
            return a;
        }
        return typeFilter;
    }
    public List<String> getRelatedLinkTo(){
        ArrayList<String> a = new ArrayList<String>();
        a.add("TestStepLinks");
        return a;
    }
    public List<String> getLinkToTypeFilter(){
        //TODO TreeLogic methods
        List<String>types = new ArrayList<String>() ,
                lt = getRelatedLinkTo();
        for(String s : lt)
            types.addAll(getLinkToTypeFilter(s));
        return types;
    }
    static final Pattern new_linkto_pattern = Pattern.compile("\\[linkto: ([^\\]]*)\\]", Pattern.MULTILINE & Pattern.DOTALL);
    static final Pattern linkto_pattern = Pattern.compile("\\[<span id=\"([a-zA-Z0-9-]+)\">.{0,10}</span>linkto: ([^\\]]*)\\]", Pattern.MULTILINE & Pattern.DOTALL);
    /**
     * Update stored linkto: constructions and attribute where
     * [linkto: ] data stored.
     * @param attrname Name of attribute where links can be placed 
     */
    @SuppressWarnings("unchecked")
    public void updateLinks(TreeNode node, String attrname, boolean removeHiddenLinks){
        boolean changed = false;
        String newValue = node.getStringAttributeValue(attrname),oldValue = newValue+"";
        if(newValue == null)
            return;
        //get change listener of this node
        INodeChangeListener changeListener = node.getChangeListener();
        TreeDB db = node.getTreeDB();
        //get links of this node
        Attribute a = node.getAttribute(TreeNode.ATTR_LINKS);
        //list of links in format UUID_UFName
        List<String> attr = null;
        TreeMap<String,String> set = null;
        if(a != null)
            attr = (List<String>)(a.getValue());
        //find stored links and 
        Matcher m = linkto_pattern.matcher(newValue);
        if(m.find()){
            set = new TreeMap<String,String>();
            String ufName = "", uid=""; 
            while(!m.hitEnd()){
                TreeNode n = null;
                try
                {
                   if(m.group(1) != null){
                       uid = m.group(1);
                       //if UUID starts with * then
                       //element is hidden (link in title)
                       if(uid.startsWith("*"))
                           uid = uid.substring(1);
                       n = db.getNode(UUID.fromString(uid));
                       //remove old listeners
                       //n.removeChangeListener(changeListener);
                   }
                }
                catch(Exception e)
                {
                }
                   ufName = m.group(2);
                   //check if all right
                   // set some properties to non-default values
                   if(n != null){
                       if(ufName.equals(n.getUserFriendlyName()))
                           set.put(uid, uid + "_" + n.getUserFriendlyName());
                   }
                m.find();
            }
        }
        //look for new links and add it
        Matcher newm = new_linkto_pattern.matcher(newValue);
        if(newm.find()){
            //if we have no links before then initiate
            //set for results
            if(set == null)
                set = new TreeMap<String,String>();
            String ufname = "", uid="";
            while(!newm.hitEnd()){
                TreeNode n = null;
                try
                {
                    ufname = newm.group(1);
                    String[] ar = db.getNodesByUFName(ufname, getLinkToTypeFilter());
                    //if we have exacly 1 node with given ufname
                    //then this link is correct
                    if(ar.length == 1)
                        n = db.getNode(UUID.fromString(ar[0]));
                    //n.removeChangeListener(changeListener);
                }
                catch(Exception e)
                {
                }
                //if we found one node then add it in
                //attribute in correct format 
                if(n != null){
                    uid = n.getUUId().toString();
                    set.put(uid, uid + "_" + ufname);
                    newValue = newValue.replace(newm.group(0), 
                            "[<span id=\"" + uid + "\">&zwnj;</span>linkto: " + ufname + "]");
                }
                newm.find();
            }
        }
        //another option - nodes that are stored in attributes
        //we need to remove listeners
        if(set == null)
           set = new TreeMap<String, String>();
        if(attr != null && attr.size() > 0){
            int num = attr.size();
            for(int i = 0; i < num; i++)
            {
                String attr_str = attr.get(i);
                String temp = attr_str.substring(0, attr_str.indexOf('_'));
                TreeNode n = null;
                //we need to save top link in this case
                //now all links that are not in attribute
                //will be deleted, but only exception is
                //"hidden" link related to title link
                if(temp.startsWith("*"))
                {
                    //if we dont want to remove hidden
                    //link then save it
                    if(!removeHiddenLinks)
                        set.put(temp, attr_str);
                    temp = temp.substring(1);
                }
                try
                {
                    //try to get node and remove old listener
                    //from it
                    n = db.getNode(UUID.fromString(temp));
                    n.removeChangeListener(changeListener);
                }
                catch(Exception e)
                {
                }
            }
        }
        //now it is time to save collected
        //links and add listeners
        int cnt = 0;
        for(String s : set.keySet()){
            try
            {
                if(attr != null)
                    if(attr.contains(set.get(s)))
                        cnt++;
                TreeNode n;
                //"hidden" links are sometimes tricky one
                if(s.startsWith("*"))
                    s = s.substring(1);
                n = db.getNode(UUID.fromString(s));
                //add listener to node
                n.addChangeListener(changeListener);
            }
            catch(Exception e)
            {
            }
        }
        if(cnt != (attr==null?0:(attr==null?0:attr.size())) || cnt != (set==null?0:(set==null?0:set.size())))
            changed = true;
        //in the case when text changed then we
        //need to update attribute
        if(!oldValue.equals(newValue)){
            node.putAttribute(new Attribute(AttributeType.STRING,attrname,newValue));
            changed = true;
        }
        //if we have no link then remove linkto attribute
        //else replace\create it
        if (set == null || set.size() == 0){
            changed = changed || (attr != null && attr.size() != 0);
            node.removeAttribute(TreeNode.ATTR_LINKS);
        }else{
            node.putAttribute(
                new Attribute(
                        AttributeType.LIST,
                        TreeNode.ATTR_LINKS, 
                        new ArrayList<String>(set.values())
                    )
                );
        }
        if(changed){
            node.saveAttributes();
            //fire event "links changed"
            fireNodeChange(node, true);
            node.fireNodeAttributeChange(TreeNode.ATTR_LINKS);
        }
    }
    public void updateLinks(TreeNode node, String attrname){
        updateLinks(node, attrname, false);
    }

}
