/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.File;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import com.unitesk.requality.tools.Index;
import com.unitesk.requality.tools.RequalityCLI;

public class TreeNode {
    public static final String ATTR_TYPE = "_type";
    public static final String ATTR_NAME = "_name";
    protected static String NumberRegex = "(-){0,1}\\d*";
    public static final String ATTR_PREDICATE = "_predicate";
    public static final String ATTR_DESCRIPTION = "_description";
    public static final String ATTR_BASE_NODES = "_baseNodes";
    public static final String ATTR_LINKS = "_links";
    public static final String NODE_GROUP = "com.unitesk.nodegroup.requality";
    //attribute for common undefined change over node
	public static final String ATTR_UNDEFINED = "_undefined";

    private UUID uuid;
    private final TreeDB db;
    protected String id;
    UUID parent;
    List<UUID> children = new ArrayList<UUID>();
    Map<String, Attribute> attributes = new LinkedHashMap<String, Attribute>();
    private boolean disposed = false;

    public String getGroup(){
    	return NODE_GROUP;
    }
    
    public TreeNode(TreeDB db, UUID uuid) {
        if (uuid == null)
            uuid = UUID.randomUUID();
        this.uuid = uuid;
        this.db = db;
        this.disposed = false;
        String indexStr = (String)getAttributeValue( TreeDB.INDEX_ATTRIBUTE );
        Attribute a = getAttribute( TreeDB.ORDER_ATTRIBUTE );
        String id = getId();
        if(indexStr == null && a == null && id != null)
            index = Index.create( id );
    }

    public void setPredicate(String pred) {
        putAttribute(new Attribute(AttributeType.STRING, ATTR_PREDICATE, pred));
    }

    public String getPredicate() {
        if (getAttributeValue(ATTR_PREDICATE) == null) {
            return "";
        }
        return (String) getAttributeValue(ATTR_PREDICATE);
    }

    public void setType(String type) {
        putAttribute(new Attribute(AttributeType.STRING, ATTR_TYPE, type));
    }

    public final UUID getUUId() {
        return uuid;
    }

    public void setId(String id) {
        db.setId(this, id);
        //update Index if is generated earlier
        String indexStr = (String)getAttributeValue( TreeDB.INDEX_ATTRIBUTE );
        Attribute a = getAttribute( TreeDB.ORDER_ATTRIBUTE );
        if(indexStr == null && a == null && getAttribute( TreeNode.ATTR_NAME ) == null){
            if(id.matches( NumberRegex )){
                index = Index.create( id );
            }else{
                index = Index.create( ((Integer)Integer.MAX_VALUE).toString() );
            }
        }
        updateTreeNodes(this, "_id");
    }
    
    public String getId() {
        return id;
    }

    /**
     * @return String name of node type
     */
    public String getType() {
        return (String) getAttributeValue(ATTR_TYPE);
    }

    public void setParent(TreeNode parent) {
        db.moveNode(this, parent);
    }

    public UUID getParentUUId() {
        return parent;
    }

    public void removeChild(TreeNode node) {
        db.removeNode(node);
    }

    public void setDescription(String description) {
        putAttribute(new Attribute(AttributeType.STRING,
                ATTR_DESCRIPTION, description));
    }
    
    public String getDescription(String prefix, String postfix, String basePath) {
        String desc = getDescription();
        if(desc == null)
        	return desc;
    	return RequalityCLI.getReportDescription(this,
        		getTreeDB().getStorage().getNodeResourcePath(getUUId()) + File.separator,
                desc, prefix, postfix, basePath);
    }

    
    public String getDescription(String basePath) {
        String result = getDescription("<div>", "</div>", basePath);
        return result;
    }
    
    public String getDescription() {
        return getDescription(false);
    }

    public String getDescription(boolean process) {
        if (attributes.get(ATTR_DESCRIPTION) != null) {
            String result = (String) attributes.get(ATTR_DESCRIPTION)
                    .getValue();
            if (process) {
                result = processResourcesPath(result);
            }
            return result;
        }
        return null;
    }

    public String processResourcesPath(String desc) {
        String resPath = getResourcesPath();
        String result = desc.replace("${node.resURL}", resPath);
        return result;
    }

    public TreeDB getTreeDB() {
        return db;
    }

    void setUUId(UUID uuid) {
        this.uuid = uuid;
    }

    public TreeNode getParent() {
        if (parent == null) {
            // no parent
            assert db.getRootNode().equals(this);
            return null;
        }
        return db.getNode(parent);
    }

    public int sizeChildren() {
        return children.size();
    }

    public TreeNode getChild(int i) {
        return db.getNode(children.get(i));
    }

    public String getQualifiedId() {
        if (getParent() == null) {
            return "/";
        }
        String parentQId = db.getNode(parent).getQualifiedId();
        if (parentQId.equals("/")) {
            return parentQId + id;
        }
        return parentQId + "/" + id;
    }

    public TreeNode findChild(String id) {
        int num = sizeChildren();
        for (int i = 0; i < num; i++) {
            TreeNode child = getChild(i);
            if (child.getId().equals(id)) {
                return child;
            }
        }
        return null;
    }

    public List<UUID> getChildrenUUIds() {
        return Collections.unmodifiableList(children);
    }

    @Override
    public String toString() {
        return id + "(" + getType() + ")";
    }

    public TreeNode createChild(String id) {
        if (findChild(id) != null)
            throw new IllegalArgumentException("element with id '" + id
                    + "' already exists");
        if (getType().equals("DBRoot")) {
            return db.createChildNode(uuid, id, "Node");
        }
        return db.createChildNode(uuid, id, getType());
    }

    public TreeNode createChild(String id, String type) {
        return db.createChildNode(uuid, id, type);
    }

    public void delete() {
        getParent().removeChild(this);
    }

    public Attribute putAttribute(Attribute a) {
        Attribute result = attributes.get(a.getKey());
        attributes.put(a.getKey(), a);
        return result;
    }

    public Attribute removeAttribute(String key) {
        return attributes.remove(key);
    }

    public Attribute getAttribute(String key) {
        return attributes.get(key);
    }

    public Object getAttributeValue(String key) {
        return getAttributeValue(key, null);
    }

    public Object getAttributeValue(String key, Object def) {
        Attribute value = attributes.get(key);
        return value != null ? value.getValue() : def;
    }

    public int getIntAttributeValue(String key, int def) {
        Attribute value = attributes.get(key);
        return value != null ? (Integer) value.getValue() : def;
    }

    public int getIntAttributeValue(String key) {
        return getIntAttributeValue(key, 0);
    }

    public float getFloatAttributeValue(String key, float def) {
        Attribute value = attributes.get(key);
        return value != null ? (Float) value.getValue() : def;
    }

    public float getFloatAttributeValue(String key) {
        return getFloatAttributeValue(key, 0);
    }

    public boolean getBooleanAttributeValue(String key, boolean def) {
        Attribute value = attributes.get(key);
        return value != null ? (Boolean) value.getValue() : def;
    }

    public boolean getBooleanAttributeValue(String key) {
        return getBooleanAttributeValue(key, false);
    }

    public String getStringAttributeValue(String key, String def) {
        Attribute value = attributes.get(key);
        return value != null ? (String) value.getValue() : def;
    }

    public String getStringAttributeValue(String key) {
        return getStringAttributeValue(key, null);
    }

    @SuppressWarnings("unchecked")
    public List<String> getListAttributeValue(String key) {
        return (List<String>) getAttributeValue(key, null);
    }

    public Set<String> getAttributeKeys() {
        return Collections.unmodifiableSet(attributes.keySet());
    }

    public Map<String, Attribute> getAttributes() {
        return Collections.unmodifiableMap(attributes);
    }

    public void saveAttributes() {
        db.saveAttributes(uuid, this.attributes.values());
        Attribute [] arr = attributes.values().toArray(new Attribute[0]);
        for(Attribute a : arr)
            fireNodeAttributeChange(a.getKey());
    }

    @SuppressWarnings("unchecked")
    public void addBaseNode(String nodeQId) {
        List<String> baseNodes = (List<String>) getAttributeValue(ATTR_BASE_NODES);
        if (baseNodes == null) {
            baseNodes = new ArrayList<String>();
        }
        baseNodes.add(nodeQId);
        putAttribute(new Attribute(AttributeType.LIST, ATTR_BASE_NODES,
                baseNodes));
    }

    @SuppressWarnings("unchecked")
    public boolean removeBaseNode(String nodeQId) {
        List<String> baseNodes = (List<String>) getAttributeValue(ATTR_BASE_NODES);
        if (baseNodes == null || !baseNodes.remove(nodeQId)) {
            return false;
        }
        putAttribute(new Attribute(AttributeType.LIST, ATTR_BASE_NODES,
                baseNodes));
        return true;
    }

    @SuppressWarnings("unchecked")
    public List<String> getBaseNodes() {
        if (getAttributeValue(ATTR_BASE_NODES) == null) {
            return Collections.EMPTY_LIST;
        }
        return Collections
                .unmodifiableList((List<String>) getAttributeValue(ATTR_BASE_NODES));
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (super.equals(obj))
            return true;
        if (!(obj instanceof TreeNode))
            return false;
        TreeNode node = (TreeNode) obj;
        if (!node.getUUId().equals(uuid))
            return false;
        if (!getQualifiedId().equals(node.getQualifiedId()))
            return false;
        if (!getAttributeKeys().equals(node.getAttributeKeys()))
            return false;
        for (String key : getAttributeKeys()) {
            if (!getAttributeValue(key).equals(node.getAttributeValue(key)))
                return false;
        }
        return true;
    }

    public TreeNode[] getSortedChildren() {
        return getSortedChildren(null, null);
    }

    public TreeNode[] getSortedChildren(String type) {
        return getSortedChildren(null, type);
    }
    
    public TreeNode[] getSortedChildren(Comparator<TreeNode> c, String type) {
        ArrayList<TreeNode> result = new ArrayList<TreeNode>();
        int num = sizeChildren();
        for (int i = 0; i < num; i++) {
            TreeNode ch = getChild(i);
            if (type == null || ch.getType().equals(type)) {
                result.add(ch);
            }
        }
        TreeNode[] r = (TreeNode[]) result.toArray(new TreeNode[0]);
        if (c != null) {
            Arrays.sort(r, c);
        } else {
            Arrays.sort(r, TreeNodeComparator.getComparator());
        }
        return r;
    }

    public String getResourcesPath() {
        try {
            String path = new File(db.getStorage().getNodeResourcePath(uuid)).toURI()
                    .toURL().toString();
            if (path.length() > 1 && path.endsWith("/")) {
                path = path.substring(0, path.length() - 1);
            }
            return path;
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }
        return null;
    }

    public NodeDesc toNodeDesc() {
        return toNodeDesc(true);
    }

    public NodeDesc toNodeDesc(Boolean deep) {
        return NodeDesc.fromNode(this, deep);
    }

    public InputStream getResourceContent(String name) {
        return db.getStorage().getResourceContents(uuid, name);
    }
    
    public String[] listResources() {
        return db.getStorage().listResources(uuid);
    }

    public void setResourceContent(String name, InputStream in) {
        db.setResourceContent(this, name, in, null);
    }

    public void setResourceContent(String name, InputStream in, Object data) {
        db.setResourceContent(this, name, in, data);
    }

    public void deleteResource(String name) {
        db.deleteResource(this, name);
    }

    public boolean isResourceExists(String name) {
        return db.isResourceExists(this, name);
    }
    
    protected Index index;

    public Index getIndex()
    {
        if( index == null )
        {
            String indexStr = (String)getAttributeValue( TreeDB.INDEX_ATTRIBUTE );
            Attribute a = getAttribute( TreeDB.ORDER_ATTRIBUTE );
            if(a != null)
                if(a.getValue() instanceof String){
                    if( indexStr == null ) indexStr = (String)a.getValue(); // old behaviour
                }else{
                    if( indexStr == null ) indexStr = ((Integer)a.getValue()).toString();
                }
            a = getAttribute( TreeNode.ATTR_NAME );
            if(indexStr == null)
                if(a != null){
                    if(((String)a.getValue()).matches( NumberRegex ))
                        indexStr = (String)a.getValue();
                }else{
                    if(getId()!=null && getId().matches( NumberRegex ))
                        indexStr = getId();
                }
            if(indexStr != null){
                index = Index.create( indexStr );
            }else{
                index = Index.create( ((Integer)Integer.MAX_VALUE).toString() );
            }
        }
        return index;
    }

    public void setIndex( Index index )
    {
        Attribute indexAttr = new Attribute( AttributeType.STRING, TreeDB.INDEX_ATTRIBUTE, index.toString() );
        putAttribute( indexAttr );
        this.index = index;
        updateTreeNodes(this, TreeDB.INDEX_ATTRIBUTE);
    }

    public boolean isDisposed() {
        return disposed;
    }

    public void setDisposed() {
        disposed = true;
    }
    
    public String getUserFriendlyName() {
        TreeNode parent = (TreeNode) getParent();
        String userFriendlyName = "";
        if (parent != null)
            userFriendlyName = parent.getUserFriendlyName();
        if (userFriendlyName.length() > 0) {
            return userFriendlyName + "/" + id;
        }
        return id;
    }
    /**
     * Update Requirement+ id\name, location update
     * @param req Requirement to update with subtree
     * @param what Attribute name to be updated(or "_id" to
     * identificator update)
     */
    public void updateTreeNodes(TreeNode req, String what) {
        TreeNode[] reqs = req.getSortedChildren();
        for (TreeNode node : reqs) {
            updateTreeNodes(node, what);
        }
        db.fireNodeChange(req, true);
        req.fireNodeAttributeChange(what);
    }
    
    private List<INodeChangeListener> listenerList;
    /**
     * Add Node Change Listener to notify later 
     * @param l node change listener to add
     */
    public void addChangeListener(INodeChangeListener l) {
        if(listenerList == null)
            listenerList = new LinkedList<INodeChangeListener>();
        synchronized (listenerList) {
            if(!listenerList.contains(l))
                listenerList.add(l);
        }
    }
    /**
     * Remove Node Change Listener 
     * @param l node change listener 
     */
    public void removeChangeListener(INodeChangeListener l) {
        if(listenerList == null)
            return;
        synchronized (listenerList) {
            if(listenerList.contains(l))
                listenerList.remove(l);
        }
    }
    /**
     * Fire Event of node deletion 
     */    
    public void fireNodeDelete() {
        if(listenerList == null)
            return;
        INodeChangeListener[] listeners = listenerList.toArray(new INodeChangeListener[0]);
        int[]del = new int[listeners.length];
        int j = 0;
        for(int i = 0;i<listeners.length; i++){
            INodeChangeListener k = listeners[i];
            try
            {
                k.deleted(this);
            }
            catch(Exception e){
                e.printStackTrace();
                del[j++] = i;
            }
        }
        synchronized (listenerList) {
            for(int i = 0; i < j; i++)
                listenerList.remove(listeners[del[i]]);
        }
    }
    /**
     * Fire Change of Node Attribute on Node Change Listeners
     * @param what attribute that was changed
     */
    public void fireNodeAttributeChange(String what) {
        if(listenerList == null)
            return;
        INodeChangeListener[] listeners = listenerList.toArray(new INodeChangeListener[0]);
        int[]del = new int[listeners.length];
        int j = 0;
        for(int i = 0;i<listeners.length; i++){
            INodeChangeListener k = listeners[i];
            try
            {
                k.attributeChange(this, what);
            }
            catch(Exception e){
                //in the case of exception in
                //listener, delete it
                e.printStackTrace();
                del[j++] = i;
            }
        }
        synchronized (listenerList) {
            for(int i = 0; i < j; i++)
                listenerList.remove(listeners[del[i]]);
        }
    }
    /**
     * Returns change listener that can observe
     * other nodes 
     */
    public INodeChangeListener getChangeListener() {
        return null;
    }
}
