/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.handlers.document;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import java.util.UUID;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.IDeleteHandler;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.nodetypes.DocFolder;
import com.unitesk.requality.nodetypes.Document;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;
import com.unitesk.requality.tools.DeepFirstTreeWalker;

public class DeleteDocumentHandler extends AbstractHandler implements IDeleteHandler {
    private static final String WARN_LOCATIONS = "Selected Document (one or more) contains marked Locations.",
    							WARN_DOCUMENTS = "Document and Document Folder deletion cannot be undone.";
    
    public String titleOverride(List<TreeNode> nodes) {
    	for(TreeNode node:nodes)
    		if(!DocFolder.TYPE_NAME.equals(node.getType()))
    			return null;
    	return "Document Folder";
	}
    
    public void recUpdateProcessed(TreeNode node, List<UUID> processed){
        processed.add(node.getUUId());
        for(UUID uid :node.getChildrenUUIds()){
            TreeNode nde = node.getTreeDB().getNode(uid);
            recUpdateProcessed(nde, processed);
        }
    }

	//start transaction and commit, please!
    private void recDelNode(TreeNode node, List<UUID> processed) {
        if (processed.contains(node.getUUId()))
            return;
        if (node instanceof Document) {
           deleteDocument((Document) node, processed);
        }

        if (node instanceof DocFolder) {
            while (node.sizeChildren() > 0) {
                recDelNode(node.getChild(0), processed);
            }
            node.delete();
        }
    }
    List<Requirement> reqs_to_delete;
    //start transaction and commit, please!
    private void deleteDocument(Document doc, List<UUID> processed) {
        removeLocationsFromRequirement(doc, processed);
        doc.delete();
    }

    private boolean folderHasDocumentWithLocation(DocFolder folder) {
        if (folder.sizeChildren() == 0) {
            return false;
        }

        for (int i = 0; i < folder.sizeChildren(); i++) {
            TreeNode node = folder.getChild(i);
            if (node instanceof DocFolder) {
                if (folderHasDocumentWithLocation((DocFolder) node)) {
                    return true;
                }
            }
            if (node instanceof Document) {
                Document doc = (Document) node;
                if (doc.sizeLocations() > 0) {
                    return true;
                }
            }
        }
        return false;
    }

    //start transaction and commit, please!
    private void removeLocationsFromRequirement(Document doc, List<UUID> processed) {
        TreeDB db = doc.getTreeDB();
        Set<String> locQIds = new HashSet<String>();
        for (int i = 0; i < doc.sizeChildren(); i++) {
            locQIds.add(doc.getChild(i).getQualifiedId());
        }

        for (TreeNode node : new DeepFirstTreeWalker(db.getNode(Requirement
                .getTypeRootQId()))) {
            if (processed.contains(node.getUUId()))
                continue;
            if (!node.getType().equals(Requirement.TYPE_NAME)) {
                continue;
            }

            ArrayList<Location> removeLocations = new ArrayList<Location>();
            Requirement req = (Requirement) node;

            for (Location loc : req.getLocations()) 
                if (locQIds.contains(loc.getQualifiedId())) {
                    removeLocations.add(loc);
                }
            for (Location loc : removeLocations) {
                req.removeLocation(loc.getQualifiedId());
                processed.add(loc.getUUId());
                if(req.getChildrenUUIds().size() == 0 && 
                		!reqs_to_delete.contains(req) && 
                		req.getLocationQIds().size() == 0 &&
                		req.getTreeDB().getTreeLogic().canBeDeleted(req))
                	reqs_to_delete.add(req);
            }
            req.saveAttributes();
        }
    }

    @Override
    public boolean isSuitableFor(TreeNode tn) {
        return tn.getType().equals(Document.TYPE_NAME) || tn.getType().equals(DocFolder.TYPE_NAME);
    }

    @Override
    public void process(List<TreeNode> nodes, List<UUID> processed, boolean multi)
            throws ExecutionException {
        int todelete = 0;
        for(TreeNode node: nodes){
            if(processed.contains(node.getUUId()))
                continue;
            if(!isSuitableFor(node) || !node.getTreeDB().getTreeLogic().canBeDeleted(node))
                continue;
            todelete++;
        }
        if(todelete == 0){
            return;
        }
        reqs_to_delete = new LinkedList<Requirement>();
        for(TreeNode node: nodes){
            if (processed.contains(node.getUUId()))
                continue;
            if (!node.getTreeDB().getTreeLogic().canBeDeleted(node))
                continue;
            if (node instanceof DocFolder) {
                List<UUID>toprocess = new ArrayList<UUID>();
                recUpdateProcessed(node, toprocess);
                recDelNode(node, processed);
                processed.addAll(toprocess);
            }
            if (node instanceof Document) {

                // TODO: document delete undo
                // node.getTreeDB().startTransaction("Delete Document");
                List<UUID>toprocess = new ArrayList<UUID>();
                recUpdateProcessed(node, toprocess);
                deleteDocument((Document) node, processed);
                processed.addAll(toprocess);
                // node.getTreeDB().commit();
            }
        }
        IUndoContext context = PlatformUI.getWorkbench().getOperationSupport().getUndoContext();
        PlatformUI.getWorkbench().getOperationSupport().getOperationHistory().dispose(context, true, true, true);
        if (reqs_to_delete.size() > 0) {
        	String strList = "";
        	for(Requirement req:reqs_to_delete)
        		strList+= "`"+req.getUserFriendlyName()+"`, ";
            if (MessageDialog.openQuestion(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(),
                    "Delete Requirement"+((reqs_to_delete.size()>1)?"s":""), 
                    		(reqs_to_delete.size() == 1 ? "Requirement `" + reqs_to_delete.get(0).getUserFriendlyName() + "` has " 
                                                            : "Requirements " + strList.substring(0,strList.length()-2) + " have ") +
                            "no locations, delete?")) {
            	TreeDB db = reqs_to_delete.get(0).getTreeDB();
            	db.startTransaction("Delete Requirement" + (reqs_to_delete.size() > 1 ? "s" : ""));
                for (Requirement req : reqs_to_delete) {
                	if(!processed.contains(req.getUUId())){
                		recUpdateProcessed(req, processed);
                		req.delete();
                	}
                }
                db.commit();
            }
        }
    }

	@Override
	public String[] getWarnings(List<TreeNode> nodes) {
		Set<String>warns = new TreeSet<String>();
		for(TreeNode node:nodes){
			if(DocFolder.TYPE_NAME.equals(node.getType())){
				warns.add(WARN_DOCUMENTS);
				if(folderHasDocumentWithLocation((DocFolder)node))
					warns.add(WARN_LOCATIONS);
			}
			if(Document.TYPE_NAME.equals(node.getType())){
				warns.add(WARN_DOCUMENTS);
				if(((Document)node).sizeLocations() > 0)
					warns.add(WARN_LOCATIONS);
			}
		}
		return warns.toArray(new String[0]);
	}

	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {
		return null;
	}
}
