/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;

import com.unitesk.requality.core.ITreeChangeListener;
import com.unitesk.requality.core.ITreeNodeFilter;
import com.unitesk.requality.core.NodeDesc;
import com.unitesk.requality.core.TreeChangeListener;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.core.TreesTracker;

public class RequalityContentProvider implements ITreeContentProvider {
    private TreeViewer viewer;
    private ArrayList<TreeDB> dbs = new ArrayList<TreeDB>();
    private ArrayList<ITreeNodeFilter> filters = new ArrayList<ITreeNodeFilter>();

    public RequalityContentProvider() {
    }

    public boolean addFilter(ITreeNodeFilter filter) {
        return filters.add(filter);
    }

    private ITreeChangeListener treeChangeListener = new TreeChangeListener() {
        private List<TreeNode> dirtyNodes = new ArrayList<TreeNode>();
        
        private void guiRefresh(final TreeNode node) {
            if (dirtyNodes.contains(node)) {
                return;
            }
            dirtyNodes.add(node);
            viewer.getControl().getDisplay().asyncExec(new Runnable() {
                @Override
                public void run() {
                    dirtyNodes.remove(node);
                    if (viewer.getControl().isDisposed()) {
                        return;
                    }
                    if (node != null
                            && (node.equals(node.getTreeDB().getRootNode()))) {
                        viewer.refresh(TreesTracker.getResourceByNode(node)
                                .getProject());
                        return;
                    }
                    if(node!=null && !node.isDisposed())
                    	viewer.refresh(node);
                    if(node == null)
                    	viewer.refresh();
                }
            });
        }

        @Override
        public void moved(TreeNode node, TreeNode oldParent) {
            guiRefresh(node.getParent());
            guiRefresh(oldParent);
        }

        @Override
        public void deleted(NodeDesc node) {
            guiRefresh(null);
        }

        @Override
        public void changed(TreeNode node) {
            guiRefresh(node.getParent());
        }

        @Override
        public void created(TreeNode node) {
            guiRefresh(node.getParent());
        }
    };

    private ITreeStateListener stateListener = new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
            dbs.add(db);
            db.addListener(treeChangeListener);
        }

        @Override
        public void delete(IProject proj) {
        }

        @Override
        public void close(TreeDB db) {
            dbs.remove(db);
        }
    };

    private boolean filterNode(TreeNode node) {
        for (ITreeNodeFilter filter : filters) {
            if (!filter.select(node)) {
                return false;
            }
        }
        return true;
    }

    @Override
    public void dispose() {
        RequalityPlugin.getTreesTracker().removeListener(stateListener);
    }

    @Override
    public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
        RequalityPlugin.getTreesTracker().removeListener(stateListener);
        if (!(viewer instanceof TreeViewer)) {
            throw new IllegalArgumentException();
        }
        this.viewer = (TreeViewer) viewer;
        for (TreeDB db : dbs) {
            db.removeListener(treeChangeListener);
        }
        dbs.clear();

        RequalityPlugin.getTreesTracker().addListener(stateListener);
        if (newInput == null) {
            return;
        }
        if (newInput instanceof TreeNode) {
            dbs.add(((TreeNode) newInput).getTreeDB());
        } else if (newInput instanceof IWorkspaceRoot) {
            dbs.addAll(Arrays.asList(RequalityPlugin.getTreesTracker()
                    .getTrees()));
        } else if (newInput instanceof Object[]) {
            dbs.add(((TreeNode) ((Object[]) newInput)[0]).getTreeDB());
        } else if (newInput instanceof TreeDB) {
            dbs.add((TreeDB) newInput);
        } else {
            throw new IllegalArgumentException();
        }
        for (TreeDB db : dbs) {
            db.addListener(treeChangeListener);
        }
        viewer.getControl().getDisplay().asyncExec(new Runnable() {
            @Override
            public void run() {
                //TODO if control is disposing, old code will throw exception
                if((RequalityContentProvider.this.viewer.getControl() != null) && 
                        !RequalityContentProvider.this.viewer.getControl().isDisposed())
                    RequalityContentProvider.this.viewer.refresh();
            }
        });
    }

    @Override
    public Object[] getElements(Object inputElement) {
        if (inputElement instanceof IWorkspaceRoot) {
            return ResourcesPlugin.getWorkspace().getRoot().getProjects();
        }
        if (inputElement instanceof TreeNode) {
            TreeNode nodeInput = (TreeNode) inputElement;
            ArrayList<TreeNode> result = new ArrayList<TreeNode>();
            for (TreeNode node : nodeInput.getSortedChildren())
                if (filterNode(node)) {
                    result.add(node);
                }
            return result.toArray(new TreeNode[0]);
        }
        if (inputElement instanceof TreeDB) {
            return new Object[] { TreesTracker.getTracker().getProject(
                    (TreeDB) inputElement) };
        }
        if (inputElement instanceof Object[]) {
            return (Object[]) inputElement;
        }
        return null;
    }

    @Override
    public Object[] getChildren(Object parentElement) {
        TreeNode parent = null;
        if (parentElement instanceof IProject) {
            IProject project = (IProject) parentElement;
            if (!project.isOpen()) {
                return new Object[0];
            }
            if (TreesTracker.getTracker().getTree(project) != null
                    && TreesTracker.getTracker().getTree(project).getRootNode() != null // Root
                                                                                        // node
                                                                                        // can
                                                                                        // be
                                                                                        // not
                                                                                        // created
                                                                                        // yet
            ) {
                parent = TreesTracker.getTracker().getTree(project)
                        .getRootNode();
            }
        }
        if (parentElement instanceof TreeNode) {
            parent = (TreeNode) parentElement;
        }
        if (parentElement instanceof TreeDB) {
            parent = ((TreeDB) parentElement).getRootNode();
        }
        if (parent == null) {
            return new Object[0];
        }

        ArrayList<TreeNode> result = new ArrayList<TreeNode>();
        for (TreeNode node : parent.getSortedChildren()) {
            if (filterNode(node)) {
                result.add(node);
            }
        }
        return result.toArray(new TreeNode[0]);
    }

    @Override
    public Object getParent(Object element) {
        if (element instanceof TreeNode) {
            TreeNode node = (TreeNode) element;
            if (node.isDisposed()) {
                return null;
            }
            if (node.getTreeDB().getRootNode().equals(node.getParent())) {
                return TreesTracker.getTracker().getProject(node.getTreeDB());
            }
            return node.getParent();
        }
        return null;
    }

    @Override
    public boolean hasChildren(Object element) {
        return getChildren(element).length > 0;
    }
}
