/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.eclipse.views.properties;

import java.util.UUID;

import org.eclipse.core.resources.IProject;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.properties.tabbed.AbstractPropertySection;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;

import com.unitesk.requality.core.INodeChangeListener;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.eclipse.core.ITreeStateListener;
import com.unitesk.requality.eclipse.core.TreesTracker;
import com.unitesk.requality.nodetypes.Location;
import com.unitesk.requality.nodetypes.Requirement;

public abstract class TreeNodePropertiesTab extends AbstractPropertySection {
    protected TreeNode node;
    boolean isDisposed = true;

    INodeChangeListener nodeChangeListener = new PropertyTabNodeChangeListener();

    private class PropertyTabNodeChangeListener implements INodeChangeListener {
    	
    	private String tabClassName;
    	{
            tabClassName = TreeNodePropertiesTab.this.getClass().getName();
        }

        public String getTabClassName() {
            return tabClassName;
        }

		@Override
		public void attributeChange(TreeNode source, String what) {
			if(source == null)
				return;
			if(node == null)
				return;
			if(source.getQualifiedId().contains(node.getQualifiedId())){
				setTreeNodeSafe(node);
				return;
			}
			setTreeNodeSafe(source);		
		}

		@Override
		public void deleted(TreeNode source) {
			if (source.getUUId().equals(uuid)) {
                node = null;
                uuid = null;
                if(page.getControl().isDisposed()) {
                    return;
                }
                if (page.getCurrentTab() != null) {
                    ISelection selection = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getSelectionService().getSelection();
                    page.selectionChanged(getPart(), selection);
                }
            }
		}
    	
    }

    ITreeStateListener treeStateListener = new ITreeStateListener() {
        @Override
        public void open(TreeDB db) {
            setTreeNodeSafe(node);
        }

        @Override
        public void close(TreeDB db) {
            setTreeNodeSafe(node);
        }

        @Override
        public void delete(IProject proj) {
        }
    };
    private UUID uuid;
    private TabbedPropertySheetPage page;

    @Override
    public void createControls(Composite parent,
            TabbedPropertySheetPage aTabbedPropertySheetPage) {
        page = aTabbedPropertySheetPage;
        super.createControls(parent, aTabbedPropertySheetPage);
        isDisposed=false;
    }

    public TreeNodePropertiesTab() {
        super();
        TreesTracker.getTracker().addListener(treeStateListener);
    }

    @Override
    public void dispose() {
        isDisposed = true;
        if(node != null)
        	node.removeChangeListener(nodeChangeListener);
        node = null;
        uuid = null;
        TreesTracker.getTracker().removeListener(treeStateListener);
        super.dispose();
    }
    protected boolean updating = false;
    private void setTreeNodeSafe(final TreeNode newNode) {
    	if(node != null)
    		node.removeChangeListener(nodeChangeListener);
        Display d = Display.getCurrent();
        if (d == null) {
            d = Display.getDefault();
        }
        updating = true;
        d.asyncExec(new Runnable() {
            @Override
            public void run() {
                if (!isDisposed) {
                    setTreeNode(newNode);
                    updating = false;
                }
            }
        });
    	if(newNode != null)
    		newNode.addChangeListener(nodeChangeListener);
    }

    abstract protected void setTreeNode(TreeNode node);

    @Override
    public void setInput(IWorkbenchPart part, ISelection selection) {
        super.setInput(part, selection);
        node = TreeNodePropertiesTab.toTreeNode(selection);
        if (node != null) {
            uuid = node.getUUId();
            setTreeNodeSafe(node);
        }
    }

    protected Requirement toRequirement(TreeNode node) {
        if (node instanceof Requirement) {
            return (Requirement) node;
        }

        if (node instanceof Location) {
            Location location = (Location) node;
            Requirement[] reqs = location.getRequirements();
            if (reqs.length > 0) {
                return reqs[0];
            }
        }
        return null;
    }

    public static TreeNode toTreeNode(ISelection sel) {
        TreeNode node = (TreeNode) ((IStructuredSelection) sel)
                .getFirstElement();
        if (node instanceof TreeNode) {
            return (TreeNode) node;
        }
        return null;
    }
}
