package com.unitesk.requality.merge;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.ranges.DocumentRange;
import org.w3c.dom.ranges.Range;

public class Chunk {

	private Node spanNode;
	private String locationText;
	private Node anchorNode;
	private String id;
	private Range textRange;

	@SuppressWarnings("serial")
	public class InvalidLocationChunkException extends Exception {

	}

	public Chunk(Node x) throws InvalidLocationChunkException {
		if (!isLocation(x))
			throw new InvalidLocationChunkException();
		spanNode = x;
		id = retrieveLocationId(x);
		anchorNode = retrieveAnchorNode(spanNode);
		locationText = anchorNode.getDomNode().getNodeValue();
		DocumentRange doc = (DocumentRange) spanNode.getDomNode()
				.getOwnerDocument();
		textRange = doc.createRange();
		textRange.setStartBefore(spanNode.getDomNode());
		textRange.setEndAfter(spanNode.getDomNode());
	}

	public Chunk(Chunk old, Node newAnchor) throws InvalidLocationChunkException {
		spanNode = old.spanNode;
		id = old.id;
		locationText = old.locationText;
		anchorNode = newAnchor;
		DocumentRange doc = (DocumentRange) anchorNode.getDomNode()
				.getOwnerDocument();
		textRange = doc.createRange();
		int startOffset = locateTextStrict(anchorNode.getText(), locationText, false);
		int endOffset = locateTextStrict(anchorNode.getText(), locationText, true);
		if(startOffset==-1) {
			startOffset = locateTextFuzzy(anchorNode.getText(), locationText, false);
			endOffset = locateTextFuzzy(anchorNode.getText(), locationText, true);
			if(startOffset==endOffset || startOffset>endOffset || endOffset-startOffset < locationText.length())
				throw new InvalidLocationChunkException();
		}
		textRange.setStart(anchorNode.getDomNode(), startOffset);
		textRange.setEnd(anchorNode.getDomNode(), endOffset);
	}

	private Node retrieveAnchorNode(Node span)
			throws InvalidLocationChunkException {
		for (Node x : span.getChilds())
			if (x.isText())
				return x;
		throw new InvalidLocationChunkException();
	}

	private int locateTextStrict(String context, String textToLocate, boolean reverse) {
		if(context.equals(textToLocate) && !reverse)
			return 0;
		else if(context.equals(textToLocate))
			return context.length();
		
		int offset = 0;
		
		if (!reverse) {
			offset = context.indexOf(textToLocate);
		} else {
			offset = context.indexOf(textToLocate)+textToLocate.length();
		}
		return offset;
	}
	
	private int locateTextFuzzy(String context, String textToLocate, boolean reverse) {
		if(Utils.jaccardSimilarity(context,textToLocate)==1.0 && !reverse)
			return 0;
		else if(Utils.jaccardSimilarity(context,textToLocate)==1.0)
			return context.length();
		
		int offset = 0;
		double[] similarityIndex = new double[context.length()];

		if (!reverse) {
			for (int i = 0; i < context.length(); i++) {
				String ctx = context.substring(i);
				similarityIndex[i] = Utils.jaccardSimilarity(ctx,
						textToLocate);
			}

			double max = similarityIndex[0];
			for (int i = 1; i < similarityIndex.length; i++) {
				if (similarityIndex[i] > max) {
					max = similarityIndex[i];
					offset = i;
				} else if (similarityIndex[i] == max && i > offset) {
					offset = i;
				}
			}
		} else {
			for (int i = context.length() - 1; i >= 0; i--) {
				String ctx = context.substring(0, i);
				similarityIndex[i] = Utils.jaccardSimilarity(ctx,
						textToLocate);
			}

			double max = similarityIndex[0];
			for (int i = 1; i < similarityIndex.length; i++) {
				if (similarityIndex[i] > max) {
					max = similarityIndex[i];
					offset = i;
				}
			}
		}
		return offset;
	}

	private String retrieveLocationId(Node x) {
		org.w3c.dom.Node attr = x.getDomNode().getAttributes()
				.getNamedItem("class");
		String attrValue = attr.getNodeValue();
		String id = attrValue.substring(attrValue.indexOf("id_") + 3);
		return id;
	}

	public void setTextRange(Range textRange) {
		this.textRange = textRange;
	}

	public Range getTextRange() {
		return this.textRange;
	}

	public String getLocationId() {
		return id;
	}

	public Node getSpanNode() {
		return spanNode;
	}

	public static boolean isLocation(Node n) {
		org.w3c.dom.Node x = n.getDomNode();
		if (!x.getNodeName().equals("span"))
			return false;
		NamedNodeMap attrs = x.getAttributes();
		if (attrs == null)
			return false;
		org.w3c.dom.Node classAtt = attrs.getNamedItem("class");
		if (classAtt == null)
			return false;
		String attValue = classAtt.getNodeValue();
		if (!attValue.contains("requality_text"))
			return false;
		return true;
	}

	public String toString() {
		return this.id + " : " + this.locationText;
	}

	public Node getAnchorNode() {
		return anchorNode;
	}

	public String getChunkText() {
		return this.locationText;
	}
}
