/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality.nodetypes;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import com.unitesk.requality.core.Attribute;
import com.unitesk.requality.core.AttributeType;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.tools.Index;

public class Requirement extends TreeNode implements Comparable<TreeNode> {
    public static final String TYPE_NAME = "Requirement";
    public static final String ATTR_NAME = "_name";
    public static final String ATTR_LOCATIONS = "_locations";
    private static final String TYPE_ROOT_QID = "/Requirements";
    public static final String ATTR_STATUS = "coverageStatus";
    public static final String ATTR_ORDER = "_order";
    public static String getTypeRootQId() {
        return TYPE_ROOT_QID;
    }

    // Constructor for eclipse instancer
    public Requirement() {
        this(null, null);
    }

    public Requirement(TreeDB db, UUID uuid) {
        super(db, uuid);
        setType(TYPE_NAME);
    }

    @SuppressWarnings("unchecked")
    public List<String> getLocationQIds() {
        Object val = getAttributeValue(ATTR_LOCATIONS);
        if (val == null) {
            return Collections.emptyList();
        }
        if (val instanceof List) {
            return Collections.unmodifiableList((List<String>) val);
        }
        throw new IllegalStateException(ATTR_LOCATIONS
                + " value must be string list");
    }

    public List<Location> getLocations() {
        List<Location> result = new ArrayList<Location>();
        for (String locQId : getLocationQIds()) {
            TreeNode locNode = getTreeDB().getNode(locQId);
            if (locNode != null && locNode instanceof Location) {
                result.add((Location) locNode);
            }
        }
        return Collections.unmodifiableList(result);
    }

    public Location getLocation(String QId) {
        if (getLocationQIds().contains(QId)) {
            return (Location) getTreeDB().getNode(QId);
        }
        return null;
    }

    @SuppressWarnings("unchecked")
    public void addLocation(String nodeQId) {
        List<String> locations = (List<String>) getAttributeValue(ATTR_LOCATIONS);
        if (locations == null) {
            locations = new ArrayList<String>();
        }
        locations.add(nodeQId);
        putAttribute(new Attribute(AttributeType.LIST, ATTR_LOCATIONS,
                locations));
        // destroy locations label cache
        for (String l : locations) {
            Location loc = (Location) getTreeDB().getNode(l);
            if (loc != null) { // Can be if node not created still
                loc.resetLabel();
            }
        }
    }

    @SuppressWarnings("unchecked")
    public boolean removeLocation(String nodeQId) {
        Location loc = (Location) getTreeDB().getNode(nodeQId);
        if (loc != null) { // Can be if node not created still
            loc.resetLabel();
        }
        List<String> locations = (List<String>) getAttributeValue(ATTR_LOCATIONS);
        if (locations == null || !locations.remove(nodeQId)) {
            return false;
        }
        if (locations.size() > 0) {
            putAttribute(new Attribute(AttributeType.LIST, ATTR_LOCATIONS,
                    locations));
        } else {
            removeAttribute(ATTR_LOCATIONS);
        }
        // destroy locations label cache
        updateLocations(this);
        return true;
    }

    public void setName(String name) {
        if (name == null || name.equals("")) {
            removeAttribute(ATTR_NAME);
        } else {
            putAttribute(new Attribute(AttributeType.STRING, ATTR_NAME, name));
        }
        //update name in cache
        getTreeDB().updateRequirementName(this);
        //fire related to name update actions
        updateTreeNodes(this, "_name");
        String indexStr = (String)getAttributeValue( TreeDB.INDEX_ATTRIBUTE );
        Attribute a = getAttribute( TreeDB.ORDER_ATTRIBUTE );
        if(indexStr == null && a == null){
            String ind = null;
            if(name != null){
                if(name.matches( NumberRegex ))
                    ind = name;
            }else{
                if(getId().matches( NumberRegex ))
                    ind = getId();
            }
            if(ind != null){
                index = Index.create( ind );
            }else{
                index = Index.create( ((Integer)Integer.MAX_VALUE).toString() );
            }
        }
    }

    @Override
    public void setId(String id) {
        if (getQualifiedId().equals(Requirement.getTypeRootQId())) {
            throw new UnsupportedOperationException(
                    "Can't rename root requirement");
        }
        super.setId(id);
        //fire id update actions
    }
    @Override
    public void updateTreeNodes(TreeNode req, String what) {
        if(req.getType().equals(TYPE_NAME))
            updateLocations((Requirement)req);
        super.updateTreeNodes(req, what);
    }

    public void updateLocations(Requirement req) {
        for (String l : req.getLocationQIds()) {
            Location loc = (Location) getTreeDB().getNode(l);
            if (loc != null) { // Can be if node not created still
                loc.resetLabel();
            }
        }
    }

    public boolean isNameDefined() {
        return getAttributeValue(ATTR_NAME) != null;
    }

    public String getNameOrId() {
        if (getAttributeValue(ATTR_NAME) == null)
            return getId();
        return (String) getAttributeValue(ATTR_NAME);
    }

    @Override
    public String toString() {
        return getId() + "(" + getUUId().toString() + ")";
    }

    @Override
    public int compareTo(TreeNode o) {
        if (o instanceof Requirement) {
            Requirement req = (Requirement) o;
            return getNameOrId().compareTo(req.getNameOrId());
        } else if (o instanceof TreeNode) {
            TreeNode req = (TreeNode) o;
            return getNameOrId().compareTo(req.getId());
        }
        throw new IllegalArgumentException();
    }

    public static String getFullName(TreeNode node, TreeDB db) {
        if (node.getParent() == null) {
            return "/";
        }
        String parentFullName = Requirement.getFullName(node.getParent(), db);
        String name;
        if(node instanceof Requirement) {
            name = ((Requirement) node).getNameOrId();
        } else {
            name = node.getId();
        }
        if (!parentFullName.equals("/")) {
            parentFullName = parentFullName + "/";
        }
        return parentFullName + name;
    }
    public String getName() {
        return (String) getAttributeValue(ATTR_NAME);
    }

    @Override
    public String getUserFriendlyName() {
        if (getId().equals("Requirements"))
            return "";
        if (getName() != null)
            return getName();
        return super.getUserFriendlyName();
    }
}
